/*
 *  @(#)aic7xxx.c 1.34 94/11/30 jda
 *
 *  Adaptec 274x/284x/294x device driver for Linux.
 *  Copyright (c) 1994 The University of Calgary Department of Computer Science.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Sources include the Adaptec 1740 driver (aha1740.c), the
 *  Ultrastor 24F driver (ultrastor.c), various Linux kernel
 *  source, the Adaptec EISA config file (!adp7771.cfg), the
 *  Adaptec AHA-2740A Series User's Guide, the Linux Kernel
 *  Hacker's Guide, Writing a SCSI Device Driver for Linux,
 *  the Adaptec 1542 driver (aha1542.c), the Adaptec EISA
 *  overlay file (adp7770.ovl), the Adaptec AHA-2740 Series
 *  Technical Reference Manual, the Adaptec AIC-7770 Data
 *  Book, the ANSI SCSI specification, the ANSI SCSI-2
 *  specification (draft 10c), ...
 *
 *  Multiple adapter card using the same IRQ level are not
 *  supported.  It doesn't make sense to configure the cards this
 *  this way from a performance standpoint.  Not to mention that
 *  the kernel would have to support two devices per registered IRQ.
 *
 *  ----------------------------------------------------------------
 *
 *  Modified to include support for wide and twin bus adapters,
 *  DMAing of SCBs, tagged queueing, bug fixes, and other rework
 *  of the code.
 *
 *  Boot time options were also added for ignoring the reversing
 *  of interrupts and not resetting the scsi bus.
 *
 *    Form:  aic7xxx=extended,reverse,no_reset
 *
 *    -- Daniel M. Eischen, deischen@iworks.ecn.uiowa.edu, 03/11/95
 */

#include <stdarg.h>
#include <asm/io.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/bios32.h>
#include <linux/delay.h>
#include <linux/pci.h>

#include "../block/blk.h"
#include "sd.h"
#include "scsi.h"
#include "hosts.h"
#include "aic7xxx.h"

/*
 * Defines for PCI bus support, testing twin bus support, DMAing of
 * SCBs, and tagged queueing.
 *
 *   o PCI bus support - this has been implemented and working since
 *     the December 1, 1994 release of this driver.  If you don't have
 *     a PCI bus and do not wish to configure your kernel with PCI
 *     support, then make sure this define is set to the cprrect
 *     define for PCI support (CONFIG_PCI) and configure your kernel
 *     without PCI support (make config).
 *
 *   o Twin bus support - this has been tested and does work.
 *
 *   o DMAing of SCBs - as of this writing, I have been unable to
 *     get this working.  FYI, all memory allocated by scsi_register
 *     returns addresses that are physical addresses.  There should
 *     be no need to convert a virtual address to a physical address.
 *
 *   o Tagged queueing - this driver is capable of tagged queueing
 *     but I am unsure as to how well the higher level driver implements
 *     tagged queueing.  Therefore, the maximum commands per lun is
 *     set to 2.  If you want to implement tagged queueing, ensure
 *     this define is not commented out.
 *
 *  Daniel M. Eischen, deischen@iworks.ecn.uiowa.edu, 03/11/95
 */

/*
 * We use our own define for PCI support in case the name ever changes.
 * It would then have to change only in one place - here.
 */

#ifdef CONFIG_PCI
#define AIC7XXX_PCI_SUPPORT
#endif


/* Uncomment this for testing twin bus support. */
#define AIC7XXX_TWIN_SUPPORT

/* Uncomment this for DMAing of SCBs. */
/* #define AIC7XXX_USE_DMA */
/* #define V_TO_P(x)	(x) */

/* Uncomment this for tagged queueing. */
#define AIC7XXX_TAGGED_QUEUEING

/* Uncomment this for extra delays and more prints */
/* #define DAN_DEBUG */

/*
 *  There should be a specific return value for this in scsi.h, but
 *  it seems that most drivers ignore it.
 */
#define DID_UNDERFLOW	DID_ERROR



/* EISA/VL-bus stuff */

#define MINSLOT		1
#define MAXSLOT		15
#define SLOTBASE(x)	((x) << 12)

#define MAXIRQ		15

/* AIC-7770 offset definitions */

#define O_MINREG(x)	((x) + 0xc00)		/* i/o range to reserve */
#define O_MAXREG(x)	((x) + 0xcbf)

/* 
 * SCSI Sequence Control (p. 3-11).  
 * Each bit, when set starts a specific SCSI sequence on the bus
 */
#define O_SCSISEQ(x)	((x) + 0xc00)
#define		TEMODEO		0x80
#define		ENSELO		0x40
#define		ENSELI		0x20
#define		ENRSELI		0x10
#define		ENAUTOATNO	0x08
#define		ENAUTOATNI	0x04
#define		ENAUTOATNP	0x02
#define		SCSIRSTO	0x01

/*
 * SCSI Transfer Control 1 Register (pp. 3-14,15).
 * Controls the SCSI module data path.
 */
#define O_SXFRCTL1(x)	((x) + 0xc02)
#define		BITBUCKET	0x80
#define		SWRAPEN		0x40
#define		ENSPCHK		0x20
#define		STIMESEL	0x18
#define		ENSTIMER	0x04
#define		ACTNEGEN	0x02
#define		STPWEN		0x01		/* Powered Termination */

/*
 * SCSI Control Signal Read Register (p. 3-15). 
 * Reads the actual state of the SCSI bus pins
 */
#define O_SCSISIGI(x)	((x) + 0xc03)
#define		CDI		0x80
#define		IOI		0x40
#define		MSGI		0x20
#define		ATNI		0x10
#define		SELI		0x08
#define		BSYI		0x04
#define		REQI		0x02
#define		ACKI		0x01

/*
 * SCSI Contol Signal Write Register (p. 3-16). 
 * Writing to this register modifies the control signals on the bus.  Only 
 * those signals that are allowed in the current mode (Initiator/Target) are
 * asserted.
 */
#define O_SCSISIGO(x)	((x) + 0xc03)
#define		CDO		0x80
#define		IOO		0x40
#define		MSGO		0x20
#define		ATNO		0x10
#define		SELO		0x08
#define		BSYO		0x04
#define		REQO		0x02
#define		ACKO		0x01

#define O_SCSIRATE(x)	((x) + 0xc04)		/* scsi rate */

/*
 * SCSI ID (p. 3-18).
 * Contains the ID of the board and the current target on the
 * selected channel
 */
#define O_SCSIID(x)	((x) + 0xc05)
#define		TID		0xf0		/* Target ID mask */
#define		OID		0x0f		/* Our ID mask */

/*
 * SCSI Status 0 (p. 3-21)
 * Contains one set of SCSI Interrupt codes
 * These are most likely of interest to the sequencer
 */
#define O_SSTAT0(x)	((x) + 0xc0b)
#define		TARGET		0x80		/* Board is a target */
#define		SELDO		0x40		/* Selection Done */
#define		SELDI		0x20		/* Board has been selected */
#define		SELINGO		0x10		/* Selection In Progress */
#define		SWRAP		0x08		/* 24bit counter wrap */
#define		SDONE		0x04		/* STCNT = 0x000000 */
#define		SPIORDY		0x02		/* SCSI PIO Ready */
#define		DMADONE		0x01		/* DMA transfer completed */

/*
 * Clear SCSI Interrupt 1 (p. 3-23)
 * Writing a 1 to a bit clears the associated SCSI Interrupt in SSTAT1.
 */ 
#define O_CLRSINT1(x)	((x) + 0xc0c)
#define		CLRSELTIMEO	0x80
#define		CLRATNO		0x40
#define		CLRSCSIRSTI	0x20
/*  UNUSED			0x10 */
#define		CLRBUSFREE	0x08
#define		CLRSCSIPERR	0x04
#define		CLRPHASECHG	0x02
#define		CLRREQINIT	0x01

/*
 * SCSI Status 1 (p. 3-24)
 * These interrupt bits are of interest to the kernel driver
 */
#define O_SSTAT1(x)	((x) + 0xc0c)
#define		SELTO		0x80
#define		ATNTARG 	0x40
#define		SCSIRSTI	0x20
#define		PHASEMIS	0x10
#define		BUSFREE		0x08
#define		SCSIPERR	0x04
#define		PHASECHG	0x02
#define		REQINIT		0x01

/*
 * SCSI Interrrupt Mode 1 (pp. 3-28,29).
 * Set bits in this register enable the corresponding
 * interrupt source.
 */
#define	O_SIMODE1(x)	((x) + 0xc11)
#define		ENSELTIMO	0x80
#define		ENATNTARG	0x40
#define		ENSCSIRST	0x20
#define		ENPHASEMIS	0x10
#define		ENBUSFREE	0x08
#define		ENSCSIPERR	0x04
#define		ENPHASECHG	0x02
#define		ENREQINIT	0x01

/*
 * Selection/Reselection ID (p. 3-31)
 * Upper four bits are the device id.  The ONEBIT is set when the re/selecting
 * device did not set its own ID.
 */
#define O_SELID(x)	((x) + 0xc19)
#define		SELID_MASK	0xf0
#define		ONEBIT		0x08
/*  UNUSED			0x07 */

/*
 * SCSI Block Control (p. 3-32)
 * Controls Bus type and channel selection.  In a twin channel configuration
 * addresses 0x00-0x1e are gated to the appropriate channel based on this
 * register.  SELWIDE allows for the coexistence of 8bit and 16bit devices
 * on a wide bus.
 */
#define O_SBLKCTL(x)	((x) + 0xc1f)
/*  UNUSED			0xc0 */
#define		AUTOFLUSHDIS	0x20		/* used for Rev C check */
/*  UNUSED			0x10 */
#define		SELBUSB		0x08
/*  UNUSED			0x04 */
#define		SELWIDE		0x02 
/*  UNUSED			0x01 */

/*
 * Sequencer Control (p. 3-33)
 * Error detection mode and speed configuration
 */
#define O_SEQCTL(x)	((x) + 0xc60)
#define		PERRORDIS	0x80
#define		PAUSEDIS	0x40
#define		FAILDIS		0x20
#define 	FASTMODE	0x10
#define		BRKADRINTEN	0x08
#define		STEP		0x04
#define		SEQRESET	0x02
#define		LOADRAM		0x01

/*
 * Sequencer RAM Data (p. 3-34)
 * Single byte window into the Scratch Ram area starting at the address
 * specified by SEQADDR0 and SEQADDR1.  To write a full word, simply write
 * four bytes in sucessesion.  The SEQADDRs will increment after the most
 * significant byte is written
 */
#define O_SEQRAM(x)	((x) + 0xc61)

/*
 * Sequencer Address Registers (p. 3-35) 
 * Only the first bit of SEQADDR1 holds addressing information
 */
#define O_SEQADDR0(x)	((x) + 0xc62)
#define O_SEQADDR1(x)	((x) + 0xc63)

#define O_ACCUM(x)	((x) + 0xc64)		/* accumulator */
#define O_BIDx(x)	((x) + 0xc80)		/* board id */

/*
 * Board Control (p. 3-43)
 */
#define O_BCTL(x)	((x) + 0xc84)
/*   RSVD			0xf0 */
#define		ACE		0x08		/* Support for external processors */
/*   RSVD			0x06 */
#define		ENABLE		0x01

#define O_BUSSPD(x)	((x) + 0xc86)		/* FIFO threshold bits ? */

/*
 * Host Control (p. 3-47) R/W
 * Overal host control of the device.  
 */
#define O_HCNTRL(x)	((x) + 0xc87)
/*    UNUSED			0x80 */
#define		POWRDN		0x40
/*    UNUSED			0x20 */
#define		SWINT		0x10
#define		IRQMS		0x08
#define		PAUSE		0x04
#define		INTEN		0x02
#define		CHIPRST		0x01
#define		REQ_PAUSE	IRQMS | PAUSE | INTEN
#define		UNPAUSE_274X	IRQMS | INTEN
#define		UNPAUSE_284X	INTEN

/*
 * SCB Pointer (p. 3-49)
 * Gate one of the four SCBs into the SCBARRAY window.
 */
#define O_SCBPTR(x)	((x) + 0xc90)

/*
 * Interrupt Status (p. 3-50)
 * Status for system interrupts
 */
#define O_INTSTAT(x)	((x) + 0xc91)
#define		SEQINT_MASK	0xf0		/* SEQINT Status Codes */
#define			BAD_PHASE	0x00
#define			SEND_REJECT	0x10
#define			NO_IDENT	0x20
#define			NO_MATCH	0x30
#define			MSG_SDTR	0x40
#define			MSG_WDTR	0x50
#define			MSG_REJECT	0x60
#define			BAD_STATUS	0x70
#define 	BRKADRINT 0x08
#define		SCSIINT	  0x04
#define		CMDCMPLT  0x02
#define		SEQINT    0x01
#define		INT_PEND  SEQINT | SCSIINT | CMDCMPLT  /* For polling */

/*
 * Hard Error (p. 3-53)
 * Reporting of catastrophic errors.  You usually cannot recover from 
 * these without a full board reset.
 */
#define O_ERROR(x)	((x) + 0xc92)		/* hard error */
/*    UNUSED			0xf0 */
#define		PARERR		0x08
#define		ILLOPCODE	0x04
#define		ILLSADDR	0x02
#define		ILLHADDR	0x01

/*
 * Clear Interrupt Status (p. 3-52)
 */
#define O_CLRINT(x)	((x) + 0xc92)
#define		CLRBRKADRINT	0x08
#define		CLRINTSTAT	0x04   /* UNDOCUMENTED - must be unpaused */
#define		CLRCMDINT 	0x02
#define		CLRSEQINT 	0x01

/*
 * SCB Auto Increment (p. 3-59)
 * Byte offset into the SCB Array and an optional bit to allow auto 
 * incrementing of the address during download and upload operations
 */
#define O_SCBCNT(x)	((x) + 0xc9a)
#define		SCBAUTO		0x80
#define		SCBCNT_MASK	0x1f

/*
 * Queue In FIFO (p. 3-60)
 * Input queue for queued SCBs (commands that the seqencer has yet to start)
 */
#define O_QINFIFO(x)	((x) + 0xc9b)

/*
 * Queue In Count (p. 3-60)
 * Number of queued SCBs
 */
#define O_QINCNT(x)	((x) + 0xc9c)

/*
 * Queue Out FIFO (p. 3-61)
 * Queue of SCBs that have completed and await the host
 */
#define O_QOUTFIFO(x)	((x) + 0xc9d)

/*
 * Queue Out Count (p. 3-61)
 * Number of queued SCBs in the Out FIFO
 */
#define O_QOUTCNT(x)	((x) + 0xc9e)

#define O_SCBARRAY(x)	((x) + 0xca0)

/* AIC-7870-only definitions */

#define O_DSPCISTATUS(x) ((x) + 0xc86)		/* ??? */

/* Scratch RAM offset definitions */

#define HA_TARG_SCRATCH(x)	((x) + 0xc20)	/* one byte per target starting at */
						/* this address for config values  */

#define HA_REJBYTE(x)	((x) + 0xc31)		/* 1st message in byte */
#define HA_MSG_LEN(x)	((x) + 0xc34)		/* pending message length */
#define HA_MSG_START(x)	((x) + 0xc35)		/* outgoing message body */
#define HA_ARG_1(x)	((x) + 0xc4a)		/* sdtr <-> rate parameters */
#define HA_RETURN_1(x)	((x) + 0xc4a)
#define		SEND_WDTR 0x80
#define		SEND_SDTR 0x80

#define HA_SIGSTATE(x)	((x) + 0xc4b)		/* value in SCSISIGO */
#define HA_SCBCOUNT(x)	((x) + 0xc52)		/* number of hardware SCBs */

#define HA_FLAGS(x)	((x) + 0xc53)		/* TWIN and WIDE bus flags */
#define		TWIN_BUS	0x01		/*   mask for TWIN bit */
#define		WIDE_BUS	0x02		/*   mask for WIDE bit */
#define		SENSE		0x10
#define		ACTIVE_MSG	0x20
#define		IDENTIFY_SEEN	0x40
#define		RESELECTING	0x80

#define HA_ACTIVE0(x)	((x) + 0xc54)		/* Active bits; targets 0-7 */
#define HA_ACTIVE1(x)	((x) + 0xc55)		/* Active bits; targets 8-15 */
#define	SAVED_TCL(x)	((x) + 0xc56)		/* Saved target, channel, LUN */

#define HA_SCSICONF(x)	((x) + 0xc5a)		/* SCSI config register */
#define HA_INTDEF(x)	((x) + 0xc5c)		/* interrupt def'n register */
#define HA_HOSTCONF(x)	((x) + 0xc5d)		/* host config def'n register */

#define MSG_ABORT	0x06
#define BUS_8_BIT	0x00
#define BUS_16_BIT	0x01
#define BUS_32_BIT	0x02


/* debugging code */

#define AIC7XXX_DEBUG

/*
 *  If a parity error occurs during a data transfer phase, run the
 *  command to completion - it's easier that way - making a note
 *  of the error condition in this location.  This then will modify
 *  a DID_OK status into a DID_PARITY one for the higher-level SCSI
 *  code.
 */
#define aic7xxx_parity(cmd)	((cmd)->SCp.Status)
#define aic7xxx_status(cmd)	((cmd)->SCp.sent_command)
#define aic7xxx_position(cmd)	((cmd)->SCp.have_data_in)

/*
 *  Since the sequencer code DMAs the scatter-gather structures
 *  directly from memory, we use this macro to assert that the
 *  kernel structure hasn't changed.
 */
#define SG_STRUCT_CHECK(sg) \
	((char *)&(sg).address - (char *)&(sg) != 0 ||	\
	 (char *)&(sg).length  - (char *)&(sg) != 8 ||	\
	 sizeof((sg).address) != 4 ||			\
	 sizeof((sg).length)  != 4 ||			\
	 sizeof(sg)	      != 12)

/*
 *  "Static" structures.  Note that these are NOT initialized
 *  to zero inside the kernel - we have to initialize them all
 *  explicitly.
 *
 *  We support a maximum of one adapter card per IRQ level (see the
 *  rationale for this above).  On an interrupt, use the IRQ as an
 *  index into aic7xxx_boards[] to locate the card information.
 */
static struct Scsi_Host *aic7xxx_boards[MAXIRQ + 1];

/*
 *  The maximum number of SCBs we could have for ANY type
 *  of card.  DON'T FORGET TO CHANGE THE SCB MASK IN THE
 *  SEQUENCER CODE IF THIS IS MODIFIED!
 */
#define AIC7XXX_MAXSCB	16

typedef enum {T_NONE, T_274X, T_284X, T_294X, T_MAX} aha_type;
typedef enum {T_NORMAL, T_TWIN, T_WIDE} aha_bus_type;


/*
 *  Define an enumerated type for the state of an SCB.
 */
typedef enum
{
  SCB_FREE = 0,
  SCB_ACTIVE = 0x1,
  SCB_ABORTED = 0x2,
  SCB_IMMED = 0x04,
  SCB_IMMED_FAIL = 0x08,
  SCB_SENSE = 0x10
} scb_state;

/* Define the bits for the control of an scb. */
#define SCB_NEEDWDTR	0x80
#define SCB_NEEDSDTR	0x40
#define SCB_TE		0x20
#define SCB_NEEDDMA	0x08
#define SCB_DIS		0x04
#define SCB_TAG_TYPE	0x03
#define 	SIMPLE_QUEUE	0x0
#define 	HEAD_QUEUE	0x1
#define 	OR_QUEUE	0x2

#define SCB_DOWNLOAD_SIZE 26
#define SCB_UPLOAD_SIZE 26
struct aic7xxx_scb {
	unsigned char control;
	unsigned char target_channel_lun;		/* 4/1/3 bits */
	unsigned char SG_segment_count;
	unsigned char SG_list_pointer[4]  __attribute__ ((packed));
	unsigned char SCSI_cmd_pointer[4]  __attribute__ ((packed));
	unsigned char SCSI_cmd_length;
	unsigned char RESERVED[2];			/* must be zero */
	unsigned char target_status;
	unsigned char residual_data_count[3];
	unsigned char residual_SG_segment_count;
	unsigned char data_pointer[4]  __attribute__ ((packed));
	unsigned char data_count[3];
	unsigned long host_scb  __attribute__ ((packed));
#if 0
	/*
	 *  No real point in transferring this to the
	 *  SCB registers.
	 */
	unsigned char RESERVED[2];
#endif
	/*------------ end of hardware supported fields ----------------*/
	struct aic7xxx_scb * next;  /* next ptr when in free list */
	Scsi_Cmnd          * cmd;   /* Scsi_Cmnd for this scb */
	scb_state            state;  /* current state of scb */
	unsigned int         position;  /* position in scb array */
};

static unsigned char generic_sense[6] = {REQUEST_SENSE, 0,0,0, 255, 0};


struct aic7xxx_host_config {
	int      irq;		/* IRQ number */
	int      base;		/* I/O base */
	int      maxscb;	/* hardware SCBs */
	int      unpause;	/* unpause value for HCNTRL */
	int      pause;		/* pause value for HCNTRL */
	int      scsi_id;	/* host SCSI ID */
	int	 scsi_id_b;	/* host SCSI ID B channel for twin cards */
	int      extended;	/* extended xlate? */
	aha_type type;		/* card type */
	aha_bus_type  bus_type; /* normal/twin/wide bus */
	unsigned char on_board;	/* flag for on-board 7870 */
};

/* Define a structure used for each host adapter, only one per IRQ. */
struct aic7xxx_host
{
	int base;					/* card base address */
	int maxscb;					/* hardware SCBs */
	int numscb;					/* current number of scbs */
	int startup;					/* intr type check */
	int extended;					/* extended xlate? */
	aha_type type;					/* card type */
	aha_bus_type  bus_type;				/* normal/twin/wide bus */
	unsigned char a_scanned;			/* 0 not scanned, 1 scanned */
	unsigned char b_scanned;			/* 0 not scanned, 1 scanned */
	volatile unsigned char unpause;			/* unpause value for HCNTRL */
	volatile unsigned char pause;			/* pause value for HCNTRL */
	volatile unsigned short needsdtr_copy;		/* default config */
	volatile unsigned short needsdtr;
	volatile unsigned short sdtr_pending;
	volatile unsigned short needwdtr_copy;		/* default config */
	volatile unsigned short needwdtr;
	volatile unsigned short wdtr_pending;
	struct aic7xxx_scb scb_array[AIC7XXX_MAXSCB];	/* copy of boards scb array */
	struct aic7xxx_scb * free_scb;			/* list of free SCBs */
};


/*
 *  NB.  This table MUST be ordered shortest period first.
 */
static struct {
	short period;
	short rate;
	char *english;
} aic7xxx_synctab[] = {
	{100,	0,	"10.0"},
	{125,	1,	"8.0"},
	{150,	2,	"6.67"},
	{175,	3,	"5.7"},
	{200,	4,	"5.0"},
	{225,	5,	"4.4"},
	{250,	6,	"4.0"},
	{275,	7,	"3.6"}
};

static int aic7xxx_synctab_max =
	sizeof(aic7xxx_synctab) / sizeof(aic7xxx_synctab[0]);

#ifdef AIC7XXX_DEBUG

	extern int vsprintf(char *, const char *, va_list);

	static
	void debug(const char *fmt, ...)
	{
		va_list ap;
		char buf[256];

		va_start(ap, fmt);
		  vsprintf(buf, fmt, ap);
		  printk(buf);
		va_end(ap);
	}

	static
	void debug_config(struct aic7xxx_host_config *p)
	{
		int ioport2, ioport3;

		static char *BRT[T_MAX][16] = {
			{ },					/* T_NONE */
			{
				"2",   "???", "???", "12",	/* T_274X */
				"???", "???", "???", "28",
				"???", "???", "???", "44",
				"???", "???", "???", "60"
			},
			{
				"2",  "4",  "8",  "12",		/* T_284X */
				"16", "20", "24", "28",
				"32", "36", "40", "44",
				"48", "52", "56", "60"
			},
			{
				"???", "???", "???", "???",	/* T_294X */
				"???", "???", "???", "???",
				"???", "???", "???", "???",
				"???", "???", "???", "???"
			}
		};
		static int DFT[4] = {
			0, 50, 75, 100
		};
		static int SST[4] = {
			256, 128, 64, 32
		};

		static char * BUSW[3] = {"", "-TWIN", "-WIDE"};

		ioport2 = inb(HA_HOSTCONF(p->base));
		ioport3 = inb(HA_SCSICONF(p->base));

		switch (p->type) {
		    case T_274X:
			printk("AHA274X%s AT EISA SLOT %d:\n",
			       BUSW[p->bus_type], p->base >> 12);
			break;
		    case T_284X:
			printk("AHA284X%s AT SLOT %d:\n",
			       BUSW[p->bus_type], p->base >> 12);
			break;
		    case T_294X:
			printk("AHA294X%s (PCI-bus):\n", BUSW[p->bus_type]);
			break;
		    default:
			panic("aic7xxx debug_config: internal error\n");
		}

		printk("    irq %d\n", p->irq);
		if (p->type != T_294X)
		{
		  printk ("    bus release time %s bclks\n"
		          "    data fifo threshold %d%%\n",
		          BRT[p->type][(ioport2 >> 2) & 0xf],
		          DFT[(ioport2 >> 6) & 0x3]);
		}
		else
		{
		  printk ("    bus release time %s bclks\n"
		          "    data fifo threshold %d%%\n",
		          BRT[p->type][(ioport2 >> 2) & 0xf],
		          DFT[(ioport3 >> 6) & 0x3]);
		}
		/* Come back and fix these later. */
		printk("    SCSI CHANNEL A:\n"
		       "        scsi id %d\n"
		       "        scsi bus parity check %sabled\n"
		       "        scsi selection timeout %d ms\n"
		       "        scsi bus reset at power-on %sabled\n",
		       ioport3 & 0x7,
		       (ioport3 & 0x20) ? "en" : "dis",
		       SST[(ioport3 >> 3) & 0x3],
		       (ioport3 & 0x40) ? "en" : "dis");

		if (p->type == T_274X) {
			printk("        scsi bus termination %sabled\n",
			       (ioport3 & 0x80) ? "en" : "dis");
		}
	}


#else

#	define debug(fmt, args...)
#	define debug_config(x)

#endif AIC7XXX_DEBUG

/*
 *  XXX - these options apply unilaterally to _all_ 274x/284x/294x
 *	  cards in the system.  This should be fixed, but then,
 *	  does anyone really have more than one in a machine?
 */
static int aic7xxx_extended = 0;	/* extended translation on? */
static int aic7xxx_no_reset = 0;	/* no resetting of SCSI bus */

void aic7xxx_setup(char *s, int *dummy)
{
	int i;
	char *p;

	static struct {
		char *name;
		int *flag;
	} options[] = {
		{"extended",	&aic7xxx_extended},
		{"no_reset",	&aic7xxx_no_reset},
		{NULL, NULL}
	};

	for (p = strtok(s, ","); p; p = strtok(NULL, ",")) {
		for (i = 0; options[i].name; i++)
			if (!strcmp(options[i].name, p))
				*(options[i].flag) = !0;
	}
}

static
void aic7xxx_getscb(int base, struct aic7xxx_scb *scb)
{
	/*
	 *  This is almost identical to aic7xxx_putscb().
	 */
	outb(SCBAUTO, O_SCBCNT(base));

	asm volatile("cld\n\t"
		     "rep\n\t"
		     "insb"
		     : /* no output */
		     :"D" (scb), "c" (SCB_UPLOAD_SIZE), "d" (O_SCBARRAY(base))
		     :"di", "cx", "dx");

	outb(0, O_SCBCNT(base));
}

static
void aic7xxx_putscb(int base, struct aic7xxx_scb *scb)
{
#ifdef AIC7XXX_USE_DMA
	/*
	 *  All we need to do, is to output the position
	 *  of the SCB in the SCBARRAY to the QINFIFO
	 *  of the host adapter.
	 */
	outb (scb->position, O_QINFIFO(base));
#else
	/*
	 *  By turning on the SCB auto increment, any reference
	 *  to the SCB I/O space postincrements the SCB address
	 *  we're looking at.  So turn this on and dump the relevant
	 *  portion of the SCB to the card.
	 */
	outb(SCBAUTO, O_SCBCNT(base));	/* SCBAUTO */

	asm volatile("cld\n\t"
		     "rep\n\t"
		     "outsb"
		     : /* no output */
		     :"S" (scb), "c" (SCB_DOWNLOAD_SIZE), "d" (O_SCBARRAY(base))
		     :"si", "cx", "dx");

	outb(0, O_SCBCNT(base));
#endif
}

#ifdef AIC7XXX_USE_DMA
static
void aic7xxx_putdmascb(int base, struct aic7xxx_scb *scb)
{
	/*
	 *  By turning on the SCB auto increment, any reference
	 *  to the SCB I/O space postincrements the SCB address
	 *  we're looking at.  So turn this on and dump the relevant
	 *  portion of the SCB to the card.
	 */
	outb(SCBAUTO, O_SCBCNT(base));	/* SCBAUTO */

	asm volatile("cld\n\t"
		     "rep\n\t"
		     "outsb"
		     : /* no output */
		     :"S" (scb), "c" (30), "d" (O_SCBARRAY(base))
		     :"si", "cx", "dx");

	outb(0, O_SCBCNT(base));
}
#endif

/*
 *  How much data should be transferred for this SCSI command?  Stop
 *  at segment sg_last if it's a scatter-gather command so we can
 *  compute underflow easily.
 */
static
unsigned aic7xxx_length(Scsi_Cmnd *cmd, int sg_last)
{
	int i, segments;
	unsigned length;
	struct scatterlist *sg;

	segments = cmd->use_sg - sg_last;
	sg = (struct scatterlist *)cmd->buffer;

	if (cmd->use_sg) {
		for (i = length = 0;
		     i < cmd->use_sg && i < segments;
		     i++)
		{
			length += sg[i].length;
		}
	} else
		length = cmd->request_bufflen;

	return(length);
}

static
void aic7xxx_to_scsirate(unsigned char *rate,
			 unsigned char transfer,
			 unsigned char offset,
                         int	       target)
{
	int i;

	for (i = 0; i < aic7xxx_synctab_max; i++) {

		if ((aic7xxx_synctab[i].period - transfer) >= 0) {
			*rate = (aic7xxx_synctab[i].rate << 4) | (offset & 0xf);
			printk("aic7xxx: target %d now synchronous at %sMb/s\n",
			       target,
			       aic7xxx_synctab[i].english);
			return;
		}
	}
	*rate = 0;
        printk("aic7xxx: target %d using asynchronous transfers\n",
           	target );
}

/*
 *  Pause the sequencer and wait for it to actually stop - this
 *  is important since the sequencer can disable pausing for critical
 *  sections.
 */
#define PAUSE_SEQUENCER(p)	\
	do {								\
		outb(p->pause, O_HCNTRL(p->base));			\
									\
		while ((inb(O_HCNTRL(p->base)) & PAUSE) == 0)		\
			;						\
	} while (0)

/*
 *  Unpause the sequencer.  Unremarkable, yet done often enough to
 *  warrant an easy way to do it.
 */
#define UNPAUSE_SEQUENCER(p)	\
	outb(p->unpause, O_HCNTRL(p->base))	/* IRQMS|INTEN */

/*
 *  See comments in aic7xxx_loadram() wrt this.
 */
#define RESTART_SEQUENCER(p)	\
	do {						\
		outb(SEQRESET | FASTMODE, O_SEQCTL(p->base));	\
							\
	} while (inb(O_SEQADDR0(p->base)) != 0 &&	\
		 inb(O_SEQADDR1(p->base)) != 0);	\
							\
	UNPAUSE_SEQUENCER(p);

/*
 *  Since we declared this using SA_INTERRUPT, interrupts should
 *  be disabled all through this function unless we say otherwise.
 */
static
void aic7xxx_isr(int irq, struct pt_regs * regs)
{
	int base, intstat;
	struct aic7xxx_host *p;
        struct aic7xxx_scb *scb;
	unsigned char active, ha_flags, transfer;
	unsigned char scsi_id, bus_width;
	unsigned char offset, rate, scratch;
	unsigned short target_mask;
	long flags;
	void *addr;
	int actual;
	int target, tcl;
	int scbptr;
	Scsi_Cmnd *cmd;
#if 0
static int_count = 0;
#endif

	p = (struct aic7xxx_host *)aic7xxx_boards[irq]->hostdata;
	base = p->base;

	/*
	 *  Check the startup flag - if no commands have been queued,
	 *  we probably have the interrupt type set wrong.  Reverse
	 *  the stored value and the active one in the host control
	 *  register.
	 */
	if (p->startup)
	{
	  switch (p->startup)
	  {
	    case 1:
	      /* Allow for 1 interrupt when the card is enabled. */
	      p->startup = 2;
	      break;

	    case 2:
	      p->pause = p->pause ^ IRQMS;
	      p->unpause = p->unpause ^ IRQMS;
	      /*
	       *  Reverse the interrupt type being careful to preserve
	       *  the current state (paused or unpaused).
	       */
	      outb ((inb(O_HCNTRL(p->base)) & PAUSE) | p->unpause,
	            O_HCNTRL(p->base));
	      printk("aic7xxx_isr: Switching interrupt trigger, hcntrl=0x%x\n",
	             inb(O_HCNTRL(p->base)));
	      p->startup = 3;
	      break;

	    default:
	      break;
	  }
	  return;
	}

	/*
	 *  Handle all the interrupt sources - especially for SCSI
	 *  interrupts, we won't get a second chance at them.
	 */
	intstat = inb(O_INTSTAT(base));

	if (intstat & BRKADRINT) {

		panic("aic7xxx_isr: brkadrint, error = 0x%x, seqaddr = 0x%x\n",
		      inb(O_ERROR(base)),
		      inb(O_SEQADDR1(base)) << 8 | inb(O_SEQADDR0(base)));
	}

	if (intstat & SEQINT) {

		/*
		 *  Although the sequencer is paused immediately on
		 *  a SEQINT, an interrupt for a SCSIINT or a CMDCMPLT
		 *  condition will have unpaused the sequencer before
		 *  this point.
		 */
		PAUSE_SEQUENCER(p);

		switch (intstat & SEQINT_MASK) {
		    case BAD_PHASE:
			panic("aic7xxx_isr: unknown scsi bus phase\n");
		    case SEND_REJECT:
			debug("aic7xxx_isr warning: "
			      "issuing message reject, 1st byte 0x%x\n",
			      inb(HA_REJBYTE(base)));
			break;
		    case NO_IDENT:
			panic("aic7xxx_isr: reconnecting target %d at seqaddr 0x%x "
			      "didn't issue IDENTIFY message\n",
			      (inb(O_SELID(base)) >> 4) & 0xf,
			      (inb(O_SEQADDR1(base)) << 8) | inb(O_SEQADDR0(base)));
			break;
		    case NO_MATCH:
			target = (inb(O_SELID(base)) >> 4) & 0xf;
			tcl = inb(O_SCBARRAY(base) + 1);
			/* Purposefully mask off the top bit of targets 8-15. */
			target_mask = 0x01 << (target & 0x07);

			debug("aic7xxx_isr: sequencer couldn't find match "
			      "for reconnecting target %d, channel %d, lun %d - "
			      "issuing ABORT\n", target, (tcl & 0x08) >> 3,
			      tcl & 0x07);
			if (tcl & 0x88) {
				/* Second channel stores its info in byte
				 * two of HA_ACTIVE
				 */
			  active = inb(HA_ACTIVE1(base));
			  active = active & ~(target_mask);
                          outb(active, HA_ACTIVE1(base));
			}
			else
			{
			  active = inb(HA_ACTIVE0(base));
			  active = active & ~(target_mask);
                          outb(active, HA_ACTIVE0(base));
			}
#ifdef AIC7XXX_USE_DMA
			outb (SCB_NEEDDMA, O_SCBARRAY(base));
#endif

			/* Check out why this use to be outb (0x80, O_CLRINT(base)) */
			/* clear the timeout */
			outb(CLRSELTIMEO, O_CLRSINT1(base));
			RESTART_SEQUENCER (p);
			break;
		    case MSG_SDTR:
                        /* 
                         * Help the sequencer to translate the negotiated
                         * transfer rate.  Transfer is 1/4 the period
                         * in ns as is returned by the sync negotiation
                         * message.  So, we must multiply by four.
                         */
                        transfer = (inb(HA_ARG_1(base)) << 2);
			offset = inb(O_ACCUM(base));
			scsi_id = inb(O_SCSIID(base)) >> 0x04;
			aic7xxx_to_scsirate(&rate, transfer, offset, scsi_id);
			if (inb(O_SBLKCTL(base)) & 0x08)
			  scsi_id = scsi_id + 8;  /* B channel */
			scratch = inb(HA_TARG_SCRATCH(base) + scsi_id);
			/* Preserve the wide transfer flag. */
			rate = rate | (scratch & 0x80);
			outb(rate, HA_TARG_SCRATCH(base) + scsi_id);
			outb(rate, O_SCSIRATE(base));

			target_mask = (0x01 << scsi_id);
			/* See if we initiated Sync Negotiation */
			if (p->sdtr_pending & target_mask)
			{ /*
			   * Negate the flag and don't send an SDTR
			   * back to the target.
			   */
			  p->needsdtr = p->needsdtr & ~target_mask;
			  p->sdtr_pending = p->sdtr_pending & ~target_mask;
			  outb(0, HA_RETURN_1(base));
			}
			else
			{ /* Send our own SDTR in reply. */
			  printk ("Sending SDTR!!\n");
			  outb(SEND_SDTR, HA_RETURN_1(base));
			}
			break;
		    case MSG_WDTR:
		      {
			bus_width = inb(O_ACCUM(base));
			scsi_id = inb(O_SCSIID(base)) >> 0x04;
			if (inb(O_SBLKCTL(base)) & 0x08)
			  scsi_id = scsi_id + 8;  /* B channel */
			printk ("Received MSG_WDTR, scsi_id = %d, "
			        "needwdtr = 0x%x\n", scsi_id, p->needwdtr);
			scratch = inb(HA_TARG_SCRATCH(base) + scsi_id);

			target_mask = (0x01 << scsi_id);
			if (p->wdtr_pending & target_mask)
			{ /*
			   * Negate the flag and don't send an WDTR
			   * back to the target, since we asked first.
			   */
			  p->needwdtr = p->needwdtr & ~target_mask;
			  p->wdtr_pending = p->wdtr_pending & ~target_mask;
			  outb(0, HA_RETURN_1(base));
			  switch (bus_width)
			  {
				case BUS_8_BIT:
				  scratch = scratch & 0x7f;
				  break;
				case BUS_16_BIT:
				  printk ("aic7xxx_isr: target %d using 16 "
				         "bit transfers\n", scsi_id);
				  scratch = scratch | 0x88;
				  scratch = scratch & 0xf8;
				  break;
			  }
			}
			else
			{ /* Send our own WDTR in reply. */
			  printk ("Will send WDTR!!\n");
			  switch (bus_width)
			  {
				case BUS_8_BIT:
				  scratch = scratch & 0x7f;
				  break;
				case BUS_32_BIT:
				  /* Negotiate 16 bits. */
				  bus_width = BUS_16_BIT;
				case BUS_16_BIT:
				  printk ("aic7xxx_isr: target %d using 16 "
				         "bit transfers\n", scsi_id);
				  scratch = scratch | 0x88;
				  scratch = scratch & 0xf8;
				  break;
			  }
			  outb (bus_width | SEND_WDTR, HA_RETURN_1(base));
			}
			outb (scratch, HA_TARG_SCRATCH(base) + scsi_id);
			outb (scratch, O_SCSIRATE(base));
			break;
		      }

		    case MSG_REJECT:
		      {
			/* 
			 * What we care about here is if we had an
			 * outstanding SDTR or WDTR message for this
			 * target.  If we did, this is a signal that
			 * the target is refusing negotiation.
			 */

			unsigned char targ_scratch, scsi_id;
			unsigned short mask;

                        scsi_id = inb(O_SCSIID(base)) >> 0x4;
			if (inb(O_SBLKCTL(base)) & 0x08)
			  scsi_id = scsi_id + 8;

			mask = (0x01 << scsi_id);

			targ_scratch = inb(HA_TARG_SCRATCH(base)
					   + scsi_id);

			if (p->needwdtr & mask) {
				/* note 8bit xfers and clear flag */
				targ_scratch = targ_scratch & 0x7f;
				p->needwdtr = p->needwdtr & ~mask;
				p->wdtr_pending = p->wdtr_pending & ~mask;
				outb(targ_scratch, HA_TARG_SCRATCH(base) 
				                 + scsi_id);
				printk("aic7xxx: target %d refusing "
				       "WIDE negotiation.  Using "
				       "8 bit transfers\n",
				       scsi_id);
			}
			else
			{
			  if (p->needsdtr & mask) {
				/* note asynch xfers and clear flag */
				targ_scratch = targ_scratch & 0xf0;
                                p->needsdtr = p->needsdtr & ~mask;
                                p->sdtr_pending = p->sdtr_pending & ~mask;
				outb(targ_scratch, HA_TARG_SCRATCH(base) 
				                 + scsi_id);
        			printk("aic7xxx: target %d refusing "
				       "syncronous negotiation.  Using "
				       "asyncronous transfers\n",
				       scsi_id);
			  }
			}
			/*
			 * Otherwise, we ignore it.
			 */
		 	break;
		      }
		    case BAD_STATUS:
			scsi_id = inb(O_SCSIID(base)) >> 0x04;
			scbptr = inb(O_SCBPTR(base));
			scb = &(p->scb_array[scbptr]);
			if ((scb->state != SCB_ACTIVE)
			    || (scb->cmd == NULL))
			{
			  printk("aic7xxx_isr: referenced scb not valid "
				 "during seqint 0x%x scb(%d) state(%x), cmd(%x)\n",
			         intstat, scbptr, scb->state, (unsigned int) scb->cmd);
			}
			else
			{
			  cmd = scb->cmd;
			  aic7xxx_getscb(base, scb);
			  aic7xxx_status(cmd) = scb->target_status;

			  cmd->result = cmd->result | scb->target_status;

			  /*
			   *  Did we underflow?  At this time, there's only
			   *  one other driver that bothers to check for this,
			   *  and cmd->underflow seems to be set rather half-
			   *  heartedly in the higher-level SCSI code.
			   */
			  actual = aic7xxx_length(cmd,
						  scb->residual_SG_segment_count);

			  actual -= ((scb->residual_data_count[2] << 16) |
				     (scb->residual_data_count[1] <<  8) |
				     (scb->residual_data_count[0]));

			  if (actual < cmd->underflow) {
				  printk("aic7xxx: target %d underflow - "
				         "wanted (at least) %u, got %u\n",
				         cmd->target, cmd->underflow, actual);

				  cmd->result = scb->target_status |
					       (DID_UNDERFLOW << 16);
			  }
			  /*
			   * This test is just here for debugging purposes.
			   * It will go away when the timeout problem is resolved.
			   */
			  switch (status_byte(scb->target_status))
			  {
			    case GOOD:
			      break;
			    case CHECK_CONDITION:
			      if ((aic7xxx_parity(cmd) == 0)
			          && !(cmd->flags & WAS_SENSE))
			      {
			        /* Update the timeout for the SCSI command. */
			        update_timeout (cmd, SENSE_TIMEOUT);

			        /* Send a sense command to the requesting target. */
			        cmd->flags = cmd->flags | WAS_SENSE;
			        memcpy ((void *) cmd->cmnd, (void *) generic_sense,
			                  sizeof(generic_sense));

			        cmd->cmnd[1] = cmd->lun << 5;	
			        cmd->cmnd[4] = sizeof(cmd->sense_buffer);

			        cmd->request_buffer = &cmd->sense_buffer;
			        cmd->request_bufflen = sizeof(cmd->sense_buffer);
			        cmd->use_sg = 0;
			        cmd->cmd_len = COMMAND_SIZE(cmd->cmnd[0]);
			        memset(scb, 0, SCB_DOWNLOAD_SIZE);
			        scb->target_channel_lun = ((cmd->target << 4) & 0xf0)
			          | (cmd->lun & 0xf);
#ifdef AIC7XXX_USE_DMA
				scb->control = SCB_NEEDDMA;
#endif
			        addr = cmd->cmnd;
			        scb->SCSI_cmd_length = cmd->cmd_len;
			        memcpy(scb->SCSI_cmd_pointer, &addr,
			               sizeof(scb->SCSI_cmd_pointer));

			        scb->SG_segment_count = 0;
			        memcpy(scb->data_pointer,
		       	               &cmd->request_buffer,
		       	               sizeof(scb->data_pointer));
			        memcpy(scb->data_count,
		       	               &cmd->request_bufflen,
		       	               sizeof(scb->data_count));

				outb(SCBAUTO, O_SCBCNT(base));	/* SCBAUTO */
				asm volatile("cld\n\t"
		     		  "rep\n\t"
		     		  "outsb"
		     		  : /* no output */
		     		  :"S" (scb), "c" (SCB_DOWNLOAD_SIZE), "d" (O_SCBARRAY(base))
		     		  :"si", "cx", "dx");
				outb(0, O_SCBCNT(base));

			        ha_flags = inb(HA_FLAGS(base));
			        outb (ha_flags | SENSE, HA_FLAGS(base));
			      }  /* first time sense, no errors */
			      else
			      { /*
			         * Do a normal command complete and have the
			         * scsi driver handle this condition.
			         */
			         cmd->flags = cmd->flags | ASKED_FOR_SENSE;
			      }
			      break;
			    case BUSY:
			      printk ("aic7xxx_isr: Target busy\n");
			      break;
			    default:
			      printk ("aic7xxx_isr: Unexpected target status 0x%x\n",
			              scb->target_status);
			      break;
			  }  /* end switch */
			}  /* end else of */
			break;
		    default:               /* unknown */
			debug("aic7xxx_isr: seqint, "
			      "intstat = 0x%x, scsisigi = 0x%x\n",
			      intstat, inb(O_SCSISIGI(base)));
			break;
		}
		outb(CLRSEQINT, O_CLRINT(base));		/* CLRSEQINT */
		UNPAUSE_SEQUENCER(p);
	}

	if (intstat & SCSIINT) {

		int status = inb(O_SSTAT1(base));

		scbptr = inb(O_SCBPTR(base));
		scb = &p->scb_array[scbptr];
		if ((scb->state != SCB_ACTIVE) || (scb->cmd == NULL))
		{
		  printk("aic7xxx_isr: no command for scb (scsiint)\n");
		  /*
		   *  Turn off the interrupt and set status
		   *  to zero, so that it falls through the
		   *  reset of the SCSIINT code.
		   */
		  outb(status, O_CLRSINT1(base));
		  UNPAUSE_SEQUENCER(p);
		  outb(CLRINTSTAT, O_CLRINT(base));	/* undocumented */
		  status = 0;
		  scb = NULL;
		}
		else
		{
		  cmd = scb->cmd;

		  /*
		   *  Only the SCSI Status 1 register has information
		   *  about exceptional conditions that we'd have a
		   *  SCSIINT about; anything in SSTAT0 will be handled
		   *  by the sequencer.  Note that there can be multiple
		   *  bits set.
		   */
		  if (status & SELTO)
		  {
			unsigned char target_mask = (1 << (cmd->target & 0x7));

			/*
			 *  Hardware selection timer has expired.  Turn
			 *  off SCSI selection sequence.
			 */
			outb(0, O_SCSISEQ(base));
			cmd->result = (DID_TIME_OUT << 16);
			/*
			 * Clear an pending messages for the timed out
			 * target and mark the target as free.
			 */
			ha_flags = inb(HA_FLAGS(base));
			outb(ha_flags & ~ACTIVE_MSG, HA_FLAGS(base));

			if (scb->target_channel_lun & 0x88)
			{
			  active = inb(HA_ACTIVE1(base));
			  active = active & ~(target_mask);
                          outb(active, HA_ACTIVE1(base));
			}
			else
			{
			  active = inb(HA_ACTIVE0(base));
			  active = active & ~(target_mask);
                          outb(active, HA_ACTIVE0(base));
			}

#ifdef AIC7XXX_USE_DMA
			outb (SCB_NEEDDMA, O_SCBARRAY(base));
#endif

			/*
			 *  Shut off the offending interrupt sources, reset
			 *  the sequencer address to zero and unpause it,
			 *  then call the high-level SCSI completion routine.
			 *
			 *  WARNING!  This is a magic sequence!  After many
			 *  hours of guesswork, turning off the SCSI interrupts
			 *  in CLRSINT? does NOT clear the SCSIINT bit in
			 *  INTSTAT.  By writing to the (undocumented, unused
			 *  according to the AIC-7770 manual) third bit of
			 *  CLRINT, you can clear INTSTAT.  But, if you do it
			 *  while the sequencer is paused, you get a BRKADRINT
			 *  with an Illegal Host Address status, so the
			 *  sequencer has to be restarted first.
			 */
			outb(CLRSELTIMEO, O_CLRSINT1(base));
			RESTART_SEQUENCER(p);

			outb(CLRINTSTAT, O_CLRINT(base));	/* undocumented */

			/*
			 *  This is a critical section, since we don't want the
			 *  queue routine mucking with the host data.
			 */
			save_flags(flags);
			cli();

			/*
			 *  Process the command after marking the scb as free
			 *  and adding it to the free list.
			 */
			scb->state = SCB_FREE;
			scb->cmd = NULL;
			scb->next = p->free_scb;	/* preserve next pointer */
			p->free_scb = scb;		/* add at head of list */

			restore_flags (flags);

			cmd->scsi_done(cmd);
#if 0
  printk("aic7xxx_isr: SELTO scb(%d) state(%x), cmd(%x)\n",
         scb->position, scb->state, (unsigned int) scb->cmd);
#endif
		  }

		  if (status & SCSIPERR)
		  {
			/*
			 *  A parity error has occurred during a data
			 *  transfer phase.  Flag it and continue.
			 */
			printk("aic7xxx: parity error on target %d, "
			       "channel %d, lun %d\n",
			       cmd->target,
			       (cmd->lun & 0x8) >> 3,
			       cmd->lun & 0x7);
			aic7xxx_parity(cmd) = DID_PARITY;

			/*
			 *  Clear interrupt and resume as above.
			 */
			outb(CLRSCSIPERR, O_CLRSINT1(base));
			UNPAUSE_SEQUENCER(p);

			outb(CLRINTSTAT, O_CLRINT(base));
			scb = NULL;
		  }

		  if (status & BUSFREE)
		  {
		    /*
		     *  A spurious bus free since selection.  This
		     *  should not happen.
		     */
#if 0
		    print ("aic7xxx_isr: unexpected busfree\n");
		    outb (CLRBUSFREE, O_CLRINT(base));  /* clear busfree */
#endif
		  }
		  else
		  {
		    if ((status & (SCSIPERR | SELTO)) == 0 && status)
		    {

			/*
			 *  We don't know what's going on.  Turn off the
			 *  interrupt source and try to continue.
			 */
			printk("aic7xxx_isr: sstat1 = 0x%x\n", status);
			outb(status, O_CLRSINT1(base));
			UNPAUSE_SEQUENCER(p);
			outb(CLRINTSTAT, O_CLRINT(base));	/* undocumented */
			scb = NULL;
		    }
		  }
		}  /* else */
	}

	if (intstat & CMDCMPLT) {

		int complete;

		/*
		 *  The sequencer will continue running when it
		 *  issues this interrupt.  There may be >1 commands
		 *  finished, so loop until we've processed them all.
		 */
		do {
			complete = inb(O_QOUTFIFO(base));

  			scb = &(p->scb_array[complete]);
			if ((scb->state != SCB_ACTIVE) || (scb->cmd == NULL))
			{
				printk("aic7xxx warning: "
				       "no command for scb %d (cmdcmplt)\n"
				       "QOUTCNT = %d, SCB state = 0x%x, CMD = 0x%x\n",
				       complete, inb(O_QOUTFIFO(base)),
				       scb->state,  (unsigned int) scb->cmd);
				outb(CLRCMDINT, O_CLRINT(base));
				continue;
			}
			cmd = scb->cmd;
			if (cmd->flags & WAS_SENSE)
			{
			  if (cmd->flags & ASKED_FOR_SENSE)
			  { /* Failed to obtain sense information */
			    cmd->result = aic7xxx_status(cmd)
					| (aic7xxx_parity(cmd) << 16);
			  }
			  else
			  { /*
			     *  Got sense information.  The higher level scsi
			     *  code checks the sense buffer in the scsi command
			     *  if the target status comes back with a mode of
			     *  CHECK_CONDITION.  Clear the sense flags from the
			     *  command and return the target status.
			     */
			    cmd->result = aic7xxx_status(cmd)
					| (aic7xxx_parity(cmd) << 16);
			    cmd->flags = cmd->flags & ~(WAS_SENSE | ASKED_FOR_SENSE);
			  }
			}
			else
			{
			  cmd->result = aic7xxx_status(cmd);
			}
#if 0
  printk ("aic7xxx_intr: (complete) state = %d, cmd = 0x%x, free = 0x%x\n",
    scb->state, (unsigned int) scb->cmd, (unsigned int) p->free_scb);
#endif
			/*
			 *  This is a critical section, since we don't want the
			 *  queue routine mucking with the host data.
			 */
			save_flags(flags);
			cli();

			scb->state = SCB_FREE;
			scb->next = p->free_scb;
			scb->cmd = NULL;
			p->free_scb = &(p->scb_array[scb->position]);

			restore_flags (flags);
#if 0
  if (scb != &p->scb_array[scb->position])
    printk ("aic7xxx_isr: (complete) address mismatch, pos %d\n", scb->position);
  printk ("aic7xxx_isr: (complete) state = %d, cmd = 0x%x, free = 0x%x\n",
    scb->state, (unsigned int) scb->cmd, (unsigned int) p->free_scb);
#endif

			cmd->scsi_done(cmd);

			/*
			 *  Clear interrupt status before checking
			 *  the output queue again.  This eliminates
			 *  a race condition whereby a command could
			 *  complete between the queue poll and the
			 *  interrupt clearing, so notification of the
			 *  command being complete never made it back
			 *  up to the kernel.
			 */
			outb(CLRCMDINT, O_CLRINT(base));

		} while (inb(O_QOUTCNT(base)));
	}

}

/*
 *  Probing for EISA boards: it looks like the first two bytes
 *  are a manufacturer code - three characters, five bits each:
 *
 *		 BYTE 0   BYTE 1   BYTE 2   BYTE 3
 *		?1111122 22233333 PPPPPPPP RRRRRRRR
 *
 *  The characters are baselined off ASCII '@', so add that value
 *  to each to get the real ASCII code for it.  The next two bytes
 *  appear to be a product and revision number, probably vendor-
 *  specific.  This is what is being searched for at each port,
 *  and what should probably correspond to the ID= field in the
 *  ECU's .cfg file for the card - if your card is not detected,
 *  make sure your signature is listed in the array.
 *
 *  The fourth byte's lowest bit seems to be an enabled/disabled
 *  flag (rest of the bits are reserved?).
 */

static
aha_type aic7xxx_probe(int slot, int s_base)
{
	int i;
	unsigned char buf[4];

	static struct {
		int n;
		unsigned char signature[sizeof(buf)];
		aha_type type;
	} S[] = {
		{4, { 0x04, 0x90, 0x77, 0x71 }, T_274X},	/* host adapter 274x */
		{4, { 0x04, 0x90, 0x77, 0x70 }, T_274X},	/* motherboard 274x  */
		{4, { 0x04, 0x90, 0x77, 0x56 }, T_284X}		/* 284x, BIOS enabled */
	};

	for (i = 0; i < sizeof(buf); i++) {
		/*
		 *  The VL-bus cards need to be primed by
		 *  writing before a signature check.
		 */
		outb(0x80 + i, s_base);
		buf[i] = inb(s_base + i);
	}

	for (i = 0; i < sizeof(S)/sizeof(S[0]); i++) {
		if (!memcmp(buf, S[i].signature, S[i].n)) {
			/*
			 *  Signature match on enabled card?
			 */
			if (inb(s_base + 4) & 1)
				return(S[i].type);
			printk("aic7xxx disabled at slot %d, ignored\n", slot);
		}
	}
	return(T_NONE);
}


static
void aic7xxx_loadram(int base)
{
	static unsigned char seqprog[] = {
		/*
		 *  Each sequencer instruction is 29 bits
		 *  long (fill in the excess with zeroes)
		 *  and has to be loaded from least -> most
		 *  significant byte, so this table has the
		 *  byte ordering reversed.
		 */
#		include "aic7xxx_seq.h"
	};

	/*
	 *  When the AIC-7770 is paused (as on chip reset), the
	 *  sequencer address can be altered and a sequencer
	 *  program can be loaded by writing it, byte by byte, to
	 *  the sequencer RAM port - the Adaptec documentation
	 *  recommends using REP OUTSB to do this, hence the inline
	 *  assembly.  Since the address autoincrements as we load
	 *  the program, reset it back to zero afterward.  Disable
	 *  sequencer RAM parity error detection while loading, and
	 *  make sure the LOADRAM bit is enabled for loading.
	 */
	outb(PERRORDIS | SEQRESET | LOADRAM, O_SEQCTL(base));

	asm volatile("cld\n\t"
		     "rep\n\t"
		     "outsb"
		     : /* no output */
		     :"S" (seqprog), "c" (sizeof(seqprog)), "d" (O_SEQRAM(base))
		     :"si", "cx", "dx");

	/*
	 *  WARNING!  This is a magic sequence!  After extensive
	 *  experimentation, it seems that you MUST turn off the
	 *  LOADRAM bit before you play with SEQADDR again, else
	 *  you will end up with parity errors being flagged on
	 *  your sequencer program.  (You would also think that
	 *  turning off LOADRAM and setting SEQRESET to reset the
	 *  address to zero would work, but you need to do it twice
	 *  for it to take effect on the address.  Timing problem?)
	 */
	outb(SEQRESET | FASTMODE, O_SEQCTL(base));
	do {
		/*
		 *  Actually, reset it until
		 *  the address shows up as
		 *  zero just to be safe..
		 */
		outb(SEQRESET | FASTMODE, O_SEQCTL(base));

	} while (inb(O_SEQADDR0(base)) != 0 && inb(O_SEQADDR1(base)) != 0);
}


static void aic7xxx_delay (int seconds)
{
	int i;

	for (i = 0; i < (seconds << 11); i = i + 1)
	  udelay (488);
}


static
int aic7xxx_register (Scsi_Host_Template *template,
		      struct aic7xxx_host_config *config)
{
	static char * board_name[T_MAX] = {"", "274x", "284x", "294x"};
	int i, base;
	unsigned char sblkctl;
	int max_targets;
	int found = 1;
	unsigned char target_settings;
	unsigned char scsi_conf;
	struct Scsi_Host *host;
	struct aic7xxx_host *p;

	base = config->base;

	/*
	 *  Lock out other contenders for our i/o space.
	 */
	request_region(O_MINREG(base), O_MAXREG(base)-O_MINREG(base), "aic7xxx");

	/*
	 * Read the bus type from the SBLKCTL register.  Set the FLAGS
	 * register in the sequencer for twin and wide bus cards.
	 */
	sblkctl = inb(O_SBLKCTL(base)) & 0x3f;  /* mask out upper two bits */
	switch (sblkctl)
	{
	    case 0:	/* narrow/normal bus */
		config->scsi_id = inb(HA_SCSICONF(base)) & 0x7;
		config->bus_type = T_NORMAL;
		outb(0, HA_FLAGS(base));
		break;
	    case 2:     /* Wide bus */
		config->scsi_id = inb(HA_SCSICONF(base) + 1) & 0xf;
                config->bus_type = T_WIDE;
		printk("aic7xxx : Enabling wide channel of %s-Wide\n",
		        board_name[config->type]);
		outb(WIDE_BUS, HA_FLAGS(base));
		break;
	    case 8:     /* Twin bus */
		config->scsi_id = inb(HA_SCSICONF(base)) & 0x7;
#ifdef AIC7XXX_TWIN_SUPPORT
		config->scsi_id_b = inb(HA_SCSICONF(base) + 1) & 0x7;
		config->bus_type = T_TWIN;
		printk("aic7xxx : Enabled channel B of %s-Twin\n",
		        board_name[config->type]);
		outb(TWIN_BUS, HA_FLAGS(base));
#else
		config->bus_type = T_NORMAL;
		printk("aic7xxx : Channel B of %s-Twin will be ignored\n",
		        board_name[config->type]);
		outb(0, HA_FLAGS(base));
#endif
		break;
	    default:
		printk("aic7xxx is an unsupported type 0x%x, please "
		       "mail dean@ims.com\n", inb(O_SBLKCTL(base)));
		outb(0, HA_FLAGS(base));
		return(0);
	}
	/*
	 *  Clear the upper two bits.  For the 294X cards, clearing the
	 *  upper two bits, will take the card out of diagnostic mode
	 *  and make the host adatper LED follow bus activity (will not
	 *  always be on).
	 */
	outb (sblkctl, O_SBLKCTL(base));

	/*
	 *  The IRQ level in i/o port 4 maps directly onto the real
	 *  IRQ number.  If it's ok, register it with the kernel.
	 *
	 *  NB. the Adaptec documentation says the IRQ number is only
	 *	in the lower four bits; the ECU information shows the
	 *	high bit being used as well.  Which is correct?
	 *
	 *  The 294X cards (PCI) get their interrupt from PCI BIOS.
	 */
	if ((config->type != T_294X)
	   && (config->irq < 9 || config->irq > 15)) {
		printk("aic7xxx uses unsupported IRQ level, ignoring\n");
		return(0);
	}
	
	/*
	 *  Print out debugging information before re-enabling
	 *  the card - a lot of registers on it can't be read
	 *  when the sequencer is active.
	 */
	debug_config(config);

	/*
	 *  Before registry, make sure that the offsets of the
	 *  struct scatterlist are what the sequencer will expect,
	 *  otherwise disable scatter-gather altogether until someone
	 *  can fix it.  This is important since the sequencer will
	 *  DMA elements of the SG array in while executing commands.
	 */
	if (template->sg_tablesize != SG_NONE) {
		struct scatterlist sg;

		if (SG_STRUCT_CHECK(sg)) {
			printk("aic7xxx warning: kernel scatter-gather "
			       "structures changed, disabling it\n");
			template->sg_tablesize = SG_NONE;
		}
	}
	
	/*
	 *  Register each "host" and fill in the returned Scsi_Host
	 *  structure as best we can.  Some of the parameters aren't
	 *  really relevant for bus types beyond ISA, and none of the
	 *  high-level SCSI code looks at it anyway.. why are the fields
	 *  there?  Also save the pointer so that we can find the
	 *  information when an IRQ is triggered.
	 *
	 */
	host = scsi_register(template, sizeof(struct aic7xxx_host));
	host->can_queue = config->maxscb;
#ifdef AIC7XXX_TAGGED_QUEUEING
	host->cmd_per_lun = 2;
#endif
	host->this_id = config->scsi_id;
	host->irq = config->irq;
	if (config->bus_type == T_WIDE)
	  host->max_id = 16;
	if (config->bus_type == T_TWIN)
	  host->max_lun = 16;

	aic7xxx_boards[config->irq] = host;
	p = (struct aic7xxx_host *)host->hostdata;

	/* Initialize the scb array by setting the state to free. */
	for (i = 0; i < AIC7XXX_MAXSCB; i = i + 1)
	{
	  p->scb_array[i].state = SCB_FREE;
	  p->scb_array[i].next = NULL;
	  p->scb_array[i].cmd = NULL;
	}

	p->a_scanned = 0;
	p->b_scanned = 0;
	p->base = config->base;
	p->maxscb = config->maxscb;
	p->numscb = 0;
	p->extended = config->extended;
	p->type = config->type;
	p->bus_type = config->bus_type;
	p->free_scb = NULL;

	/*
	 *  The interrupt trigger is different depending on
	 *  whether the card is EISA or VL-bus - sometimes.  The
	 *  startup variable will be cleared once the first
	 *  command is queued, and is checked in the isr to try
	 *  and detect when the interrupt type is set incorrectly,
	 *  immediately triggering spurious interrupts.  This is
	 *  now just set on a per-card-type basis.
	 *
	 *  This should not occur anymore because the interrupt
	 *  type was taken from the BIOS settings.  But we leave
	 *  it this way until we know this is the case.  We also
	 *  allow for one interrupt without a command being sent,
	 *  in case the host adapter gives one interrupt upon
	 *  being enabled.  A second interrupt will reverse the
	 *  interrupt trigger.
	 */
	p->unpause = config->unpause;
	p->pause = config->pause;
	p->startup = 1;

	/*
	 *  Register IRQ with the kernel _after_ the host information
	 *  is set up, in case we take an interrupt right away, due to
	 *  the interrupt type being set wrong.
	 */
	if (request_irq(config->irq, aic7xxx_isr, SA_INTERRUPT, "aic7xxx")) {
		printk("aic7xxx couldn't register irq %d, ignoring\n",
		       config->irq);
		return(0);
	}

	/*
	 *  Load the sequencer program, then re-enable the board -
	 *  resetting the AIC-7770 disables it, leaving the lights
	 *  on with nobody home.  On the PCI bus you *may* be home,
	 *  but then your mailing address is dynamically assigned
	 *  so no one can find you anyway :-)
	 */
        printk("aic7xxx: Downloading sequencer code..");
	aic7xxx_loadram(base);

	/* Set Fast Mode and Enable the board */
	outb (FASTMODE, O_SEQCTL(base));

	if (p->type != T_294X)
	  outb(ENABLE, O_BCTL(base));

	printk ("done.\n");


	/* Set the SCSI Id, SXFRCTL1, and SIMODE1, for both channels */
	if (p->bus_type == T_TWIN)
	{
	  /* 
	   * The device is gated to channel B after a chip reset,
	   * so set those values first.
	   */
	  outb (config->scsi_id_b, O_SCSIID(base));
	  scsi_conf = inb (HA_SCSICONF(base) + 1) & (ENSPCHK|STIMESEL);
	  scsi_conf = scsi_conf | ENSTIMER | ACTNEGEN | STPWEN;
	  outb(scsi_conf, O_SXFRCTL1(base));
	  outb (ENSELTIMO|ENSCSIPERR, O_SIMODE1(base));
	  /* Select Channel A */
	  outb(0, O_SBLKCTL(base));
	}
	outb (config->scsi_id, O_SCSIID(base));
	scsi_conf = inb (HA_SCSICONF(base)) & (ENSPCHK|STIMESEL);
	outb (scsi_conf|ENSTIMER|ACTNEGEN|STPWEN, O_SXFRCTL1(base));
	outb (ENSELTIMO|ENSCSIPERR, O_SIMODE1(base));

	/* Look at the information that board initialization or the board
	 * BIOS has left us.  In the lower four bits of each target's
	 * scratch space any value other than 0 indicates that we should
	 * initiate synchronous transfers.  If it's zero, the user or the
	 * BIOS has decided to disable synchronous negotiation to that
	 * target so we don't activate the needsdtr flag.
	 */
        p->needsdtr_copy = 0;          
        p->sdtr_pending = 0;          
        p->needwdtr_copy = 0;          
        p->wdtr_pending = 0;
	if (p->bus_type == T_NORMAL)
	  max_targets = 8;
	else
	  max_targets = 16;
	for (i = 0; i < max_targets; i = i + 1)
	{
	  target_settings = inb(HA_TARG_SCRATCH(base) + i);
	  if (target_settings & 0x0f)
	  {
	    p->needsdtr_copy = p->needsdtr_copy | (0x01 << i);
	    target_settings = target_settings | 0x0f;
	  }
	  /*
	   * If we are not wide, forget WDTR.  This makes the driver
	   * work on some cards that don't leave these fields cleared
	   * when BIOS is not installed.
	   */
	  if ((target_settings & 0x80) && (p->bus_type == T_WIDE))
	  {
	    p->needwdtr_copy = p->needwdtr_copy | (0x01 << i);
	    target_settings = target_settings & 0x7f;
	  }
	  outb(target_settings, (HA_TARG_SCRATCH(base) + i));
	}

        p->needsdtr = p->needsdtr_copy;          
        p->needwdtr = p->needwdtr_copy;          
        printk("NeedSdtr = 0x%x, 0x%x\n", p->needsdtr_copy, p->needsdtr);
        printk("NeedWdtr = 0x%x, 0x%x\n", p->needwdtr_copy, p->needwdtr);

	/*
	 *  For reconnecting targets, the sequencer code needs to
	 *  know how many SCBs it has to search through.
	 */
	outb(config->maxscb, HA_SCBCOUNT(base));

	/* Clear the active flags - no targets are busy. */
	outb(0, HA_ACTIVE0(base));
	outb(0, HA_ACTIVE1(base));

	/*
	 * Reset the SCSI bus.  Is this necessary?
	 *   To answer - No.  My Plextor PX-43CH CD-ROM does not respond
	 *   to a Test Unit Ready when a wide cable is installed with a
	 *   a wide device and the reset is performed.  If the wide cable
	 *   is disconnected and Linux boots from a floppy or non-wide
	 *   disk, then the CD-ROM does respond to the Test Unit Ready
	 *   after a reset.  The CD-ROM responds to the Test Unit Ready
	 *   with and without the wide cable/disk drive if the reset is
	 *   NOT performed.  Many hours were spent trying to figure this
	 *   one out.
	 *
	 * By delaying after the reset to let the devices settle, my
	 * CD-ROM does not hang.  I have not looked into other methods
	 * of delaying.
	 */

	if (!aic7xxx_no_reset)
	{
	  printk ("Resetting the SCSI bus...\n");
	  outb(SCSIRSTO, O_SCSISEQ(base));
	  udelay(1000);
	  outb(0, O_SCSISEQ(base));
	  aic7xxx_delay (15);
	}

	/*
	 *  Unpause the sequencer before returning and enable
	 *  interrupts - we shouldn't get any until the first
	 *  command is sent to us by the high-level SCSI code.
	 */

	UNPAUSE_SEQUENCER(p);
	return(found);
}


#ifdef AIC7XXX_PCI_SUPPORT
static
int aic7xxx_7870_register (Scsi_Host_Template *template,
			   unsigned char bus,
			   unsigned char device_fn,
			   unsigned char on_board)
{
	struct aic7xxx_host_config config;
	int error;
	unsigned long io_port;
	unsigned char irq;

	config.type = T_294X;
	config.on_board = on_board;

	/*
	 *  Read esundry information from PCI BIOS.
	 */
	error = pcibios_read_config_dword(bus,
					  device_fn,
					  PCI_BASE_ADDRESS_0,
					  &io_port);
	if (error) {
		panic("aha294x_config: error 0x%x reading i/o port\n",
		      error);
	}

	error = pcibios_read_config_byte(bus,
					 device_fn,
					 PCI_INTERRUPT_LINE,
					 &irq);
	if (error) {
		panic("aha294x_config: error 0x%x reading irq\n",
		      error);
	}

	/*
	 *  Make the base I/O register look like EISA and VL-bus.
	 */
	config.base = io_port - 0xc01;

#if 1
  printk ("aic7xxx: aic7870 hcntrl=0x%x\n", inb(O_HCNTRL(config.base)));
#endif
	/*
	 *  Give the AIC-7XXX a reset before reading any of its
	 *  its registers - this forces a pause of the sequencer
	 *  and returns everything to default values.
	 *
	 *  Use the BIOS settings to determine the interrupt
	 *  trigger type (level or edge) and use this value
	 *  for pausing and unpausing the sequencer.
	 *
	 *  For some 274X boards, we must clear the CHIPRST bit
	 *  and pause the sequencer.  For some reason, this makes
	 *  the driver work.
	 */
	config.unpause = (inb(O_HCNTRL(config.base)) & IRQMS) | INTEN;
	config.pause = config.unpause | PAUSE;
	outb(config.pause | CHIPRST, O_HCNTRL(config.base));
	aic7xxx_delay (1);
	outb(config.pause, O_HCNTRL(config.base));

	config.irq = irq;
	config.maxscb = 16;

	/*
	 *  XXX - these are values that I don't know how to query
	 *	  the hardware for, so for now the SCSI host ID is
	 *	  hardwired to 7, and the "extended translation"
	 *	  flag is taken from boot-time flags.
	 */
	config.scsi_id = 7;
	config.extended = aic7xxx_extended;

	/*
	 *  XXX - force data fifo threshold to 100%.  Why does this
	 *	  need to be done?
	 */
#	define	DFTHRESH	0xc0

	outb(inb(O_DSPCISTATUS(config.base)) | DFTHRESH, O_DSPCISTATUS(config.base));
	outb(config.scsi_id | DFTHRESH, HA_SCSICONF(config.base));
	/* In case we are a wide card, place scsi ID in second conf byte. */
        outb(config.scsi_id, (HA_SCSICONF(config.base) + 1)); 

	/*
	 *  A reminder until this can be detected automatically.
	 */
	printk("aha294x: extended translation %sabled\n",
		config.extended ? "en" : "dis");

	/*
	 *  Finish the setup and registration by calling the
	 *  generic routine.
	 */
	return (aic7xxx_register (template, &config));
}
#endif


static
int aic7xxx_7770_register (Scsi_Host_Template *template,
 			   aha_type type,
			   int      base)
{
	struct aic7xxx_host_config config;
	unsigned char sblkctl_reg;

	config.type = type;
	config.base = base;

#if 1
  printk ("aic7xxx: aic7770 hcntrl=0x%x\n", inb(O_HCNTRL(config.base)));
#endif
	/*
	 *  Give the AIC-7XXX a reset before reading any of its
	 *  its registers - this forces a pause of the sequencer
	 *  and returns everything to default values.
	 *
	 *  Use the BIOS settings to determine the interrupt
	 *  trigger type (level or edge) and use this value
	 *  for pausing and unpausing the sequencer.
	 *
	 *  For some 274X boards, we must clear the CHIPRST bit
	 *  and pause the sequencer.  For some reason, this makes
	 *  the driver work.
	 */
	config.unpause = (inb(O_HCNTRL(config.base)) & IRQMS) | INTEN;
	config.pause = config.unpause | PAUSE;
	outb(config.pause | CHIPRST, O_HCNTRL(config.base));
	aic7xxx_delay (1);
	outb(config.pause, O_HCNTRL(config.base));


	/* Get the IRQ */
        config.irq = inb(HA_INTDEF(config.base)) & 0xf;

	/*
	 * A reminder: extended translation cannot yet be
	 *             automatically detected.
	 */
	if (config.type == T_274X)
	{
	  config.extended = aic7xxx_extended;
	  printk("aha274x: extended translation %sabled\n",
	         config.extended ? "en" : "dis");
	  printk ("aha274x: using 0x%x for pausing of sequencer.\n", config.unpause);
	}
	else
	{
	  config.extended = aic7xxx_extended;
	  printk("aha284x: extended translation %sabled\n",
	         config.extended ? "en" : "dis");
	  printk ("aha284x: using 0x%x for pausing of sequencer.\n", config.unpause);
	}

	/*
	 * Check for Rev C or E boards.  Rev E boards can have
	 * 16 SCBs, while the Rev C boards are limited to 4 SCBs.
	 *
	 * The Rev E boards have a read/write autoflush bit in the
	 * SBLKCTL registor, while in the Rev C boards it is read only.
	 */
	sblkctl_reg = inb(O_SBLKCTL(config.base)) ^ AUTOFLUSHDIS;
	outb(sblkctl_reg, O_SBLKCTL(config.base));
	if (inb(O_SBLKCTL(config.base)) == sblkctl_reg)
	{  /* We detected a Rev E board. */
	  printk ("aic7770: Rev E and subsequent; using 16 SCBs :-)\n");
	  outb (sblkctl_reg ^ AUTOFLUSHDIS, O_SBLKCTL(config.base));
	  config.maxscb = 16;
	}
	else
	{
	  printk ("aic7770: Rev C and previous; using 4 SCBs :-(\n");
	  config.maxscb = 4;
	}
	/*
	 *  Finish the setup and registration by calling the
	 *  generic routine.
	 */
	return (aic7xxx_register (template, &config));
}


int aic7xxx_detect(Scsi_Host_Template *template)
{
	aha_type type;
	int found = 0, slot, base;

/* This should be defined in pci.h */
#define PCI_DEVICE_ID_ADAPTEC_2940_MB    0x7078

	/*
	 *  EISA/VL-bus card signature probe.
	 */
	for (slot = MINSLOT; slot <= MAXSLOT; slot++) {

		base = SLOTBASE(slot);
		
		if (check_region(O_MINREG(base),
				 O_MAXREG(base)-O_MINREG(base)))
		{
			/*
			 *  Some other driver has staked a
			 *  claim to this i/o region already.
			 */
			continue;
		}

		type = aic7xxx_probe(slot, O_BIDx(base));

		if (type != T_NONE) {
			/*
			 *  We "find" a 274x if we locate the card
			 *  signature and we can set it up and register
			 *  it with the kernel without incident.
			 */
			found += aic7xxx_7770_register(template, type, base);
		}
	}

#ifdef AIC7XXX_PCI_SUPPORT
	/*
	 *  PCI-bus probe.
	 */
	if (pcibios_present()) {
		/* Look for an on-board controller. */
		int index = 0;
		unsigned char bus, device_fn;

		while (!pcibios_find_device(PCI_VENDOR_ID_ADAPTEC,
					    PCI_DEVICE_ID_ADAPTEC_2940_MB,
					    index,
					    &bus,
					    &device_fn))
		{
			found += aic7xxx_7870_register(template, bus,
						       device_fn, 1);
			index += 1;
		}
		/* Look for controller not on motherboard. */
		index = 0;
		while (!pcibios_find_device(PCI_VENDOR_ID_ADAPTEC,
					    PCI_DEVICE_ID_ADAPTEC_2940,
					    index,
					    &bus,
					    &device_fn))
		{
			found += aic7xxx_7870_register(template, bus,
						       device_fn, 0);
			index += 1;
		}
	}
#endif

	template->name = (char *)aic7xxx_info(NULL);
	return(found);
}


const char *aic7xxx_info(struct Scsi_Host *notused)
{
	return("Adaptec AHA274x/284x/294x (EISA/VL-bus/PCI -> Fast SCSI) "
	       AIC7XXX_SEQ_VERSION "/"
	       AIC7XXX_H_VERSION "/"
	       "1.34");
}


static
void aic7xxx_buildscb(struct aic7xxx_host *p,
		      Scsi_Cmnd *cmd,
		      struct aic7xxx_scb *scb)
{
	void *addr;
	unsigned length;
	unsigned short mask;

   	/*
	 *  Setup the control byte if we need negotiation and have not
	 *  already requested it.  Ensure that we only do one negotiation
	 *  at a time including tag enable.  This way, if we get a reject
	 *  message, we know which message was rejected.  Also, it seems
	 *  that for some reason, we cannot perform tag enable at the
	 *  same time as wide or synchronous negotiation.
	 */

	mask = (0x01 << cmd->target);
        if ((p->needwdtr & mask) && !(p->wdtr_pending & mask))
	{
	  p->wdtr_pending = p->wdtr_pending | mask; 
          scb->control = scb->control | SCB_NEEDWDTR;
#if 0
  printk ("Sending WDTR request to target %d.\n", cmd->target);
#endif
	}
	else
	{
          if ((p->needsdtr & mask) && !(p->sdtr_pending & mask))
	  {
	    p->sdtr_pending = p->sdtr_pending | mask; 
            scb->control = scb->control | SCB_NEEDSDTR;
#if 0
  printk   ("Sending SDTR request to target %d.\n", cmd->target);
#endif
	  }

#ifdef AIC7XXX_TAGGED_QUEUEING
	  else
	  {
	    if (cmd->device->tagged_queue)
	    {
	      cmd->device->current_tag = 1;  /* enable tagging */
	      cmd->tag = cmd->device->current_tag;
	      cmd->device->current_tag = cmd->device->current_tag + 1;
	      scb->control = scb->control | SCB_TE;
	    }
	  }
#endif
	}


#if 0
  printk ("aic7xxx_queue: target %d, cmd 0x%x (size %u), wdtr 0x%x, mask 0x%x\n",
    cmd->target, cmd->cmnd[0], cmd->cmd_len, p->needwdtr, mask);
#endif
	/*
	 *  NB. channel selection is taken from the 
	 */
	scb->target_channel_lun = ((cmd->target << 4) & 0xf0) | (cmd->lun & 0xf);

	/*
	 *  The interpretation of request_buffer and request_bufflen
	 *  changes depending on whether or not use_sg is zero; a
	 *  non-zero use_sg indicates the number of elements in the
	 *  scatter-gather array.
	 *
	 *  The AIC-7770 can't support transfers of any sort larger
	 *  than 2^24 (three-byte count) without backflips.  For what
	 *  the kernel is doing, this shouldn't occur.  I hope.
	 */
	length = aic7xxx_length(cmd, 0);

	if (length > 0xffffff) {
		panic("aic7xxx_buildscb: can't transfer > 2^24 - 1 bytes\n");
	}

	/*
	 *  XXX - this relies on the host data being stored in a
	 *	  little-endian format.
	 *
	 *  I'm not sure how the scatterlist is setup and what is
	 *  and what is not physical memory.  I think that at least
	 *  the address of the command needs to be converted to
	 *  physical memory.  If the scatterlist needs to be converted,
	 *  look at the aic7xxx_length function.  A copy of that with
	 *  some slight modifications to store the scatterlist in
	 *  the scb->dma_segs array should work.  I leave this
	 *  unimplemented as of now.
	 */

	addr = cmd->cmnd;
	scb->SCSI_cmd_length = cmd->cmd_len;
	memcpy(scb->SCSI_cmd_pointer, &addr, sizeof(scb->SCSI_cmd_pointer));

	if (cmd->use_sg) {
#if 0
		debug("aic7xxx_buildscb: SG used, %d segments, length %u\n",
		      cmd->use_sg,
		      length);
#endif
		scb->SG_segment_count = cmd->use_sg;
		memcpy(scb->SG_list_pointer,
		       &cmd->request_buffer,
		       sizeof(scb->SG_list_pointer));
	} else {
		scb->SG_segment_count = 0;
		memcpy(scb->data_pointer,
		       &cmd->request_buffer,
		       sizeof(scb->data_pointer));
		memcpy(scb->data_count,
		       &cmd->request_bufflen,
		       sizeof(scb->data_count));
	}
}

int aic7xxx_queue(Scsi_Cmnd *cmd, void (*fn)(Scsi_Cmnd *))
{
	long flags;
	int old_scbptr;
	int position;
	struct aic7xxx_host *p;
	struct aic7xxx_scb *scb;
#ifdef AIC7XXX_USE_DMA
	void *addr;
	unsigned char curscb;
#endif


	p = (struct aic7xxx_host *)cmd->host->hostdata;

	/* Check to see if channel was scanned. */
	if (!p->a_scanned && (cmd->lun < 0x8))
	{
	  printk ("aic7xxx: Scanning channel A for devices.\n");
	  p->a_scanned = 1;
	}
	else
	{
	  if (!p->b_scanned && (cmd->lun & 0x8))
	  {
	    printk ("aic7xxx: Scanning channel B for devices.\n");
	    p->b_scanned = 1;
	  }
	}

#if 0
  if (p->startup)
	debug("aic7xxx_queue: cmd 0x%x (size %u), target %d, channel %d, lun %d\n",
	      cmd->cmnd[0],
	      cmd->cmd_len,
	      cmd->target,
	      (cmd->lun & 0x8) >> 3,
	      cmd->lun & 0x7);
#endif

	/*
	 *  This is a critical section, since we don't want the
	 *  interrupt routine mucking with the host data or the
	 *  card.  Since the kernel documentation is vague on
	 *  whether or not we are in a cli/sti pair already, save
	 *  the flags to be on the safe side.
	 */
	save_flags(flags);
	cli();

	/*
	 *  Clear the startup flag - we can now legitimately
	 *  expect interrupts.
	 */
	p->startup = 0;

	/*
	 *  Find a free slot in the SCB array to load this command
	 *  into.  Since can_queue is set to the maximum number of
	 *  SCBs for the card, we should always find one.
	 *
	 *  First try to find an scb in the free list.  If there are
	 *  none in the free list, then check the current number of
	 *  of scbs and take an unused one from the scb array.
	 */
	scb = p->free_scb;
	if (scb != NULL)
	{  /* found one in the free list */
	  p->free_scb = scb->next;   /* remove and update head of list */
	  scb = &p->scb_array[scb->position];
	  position = scb->position;  /* preserve the position */
	  memset(scb, 0, sizeof(*scb));
	  scb->state = SCB_ACTIVE;
	  scb->position = position;
	}
	else
	{
	  if (p->numscb >= p->maxscb)
	    panic ("aic7xxx_queue: couldn't find a free scb\n");
	  else
	  {
	    /*
	     *  Initialize the scb within the scb array.  The
	     *  position within the array is the position on
	     *  the board that it will be loaded.
	     */
	    scb = &(p->scb_array[p->numscb]);
	    memset(scb, 0, sizeof(*scb));
	    scb->position = p->numscb;
	    p->numscb = p->numscb + 1;
	    scb->state = SCB_ACTIVE;
#ifdef AIC7XXX_USE_DMA
	    addr = &scb;
	    memcpy(scb->host_scb, &addr, sizeof(scb));
	    scb->control = SCB_NEEDDMA;
	    PAUSE_SEQUENCER (p);
	    curscb = inb (O_SCBPTR(p->base));
	    outb (scb->position, O_SCBPTR(p->base));
	    aic7xxx_putdmascb (p->base, scb);
	    outb (curscb, O_SCBPTR(p->base));
	    UNPAUSE_SEQUENCER (p);
	    scb->control = 0;
#endif
	  }
	}
	scb->cmd = cmd;
	aic7xxx_position(cmd) = scb->position;
	/*
	 *  Construct the SCB beforehand, so the sequencer is
	 *  paused a minimal amount of time.
	 */
	aic7xxx_buildscb(p, cmd, scb);

#if 0
  if (scb != &p->scb_array[scb->position])
    printk ("aic7xxx_queue: address of scb by position does not match scb address\n");
  printk ("aic7xxx_queue: SCB pos=%d, cmdptr=0x%x, state=%d, freescb=0x%x\n",
          scb->position, (unsigned int) scb->cmd,
          scb->state, (unsigned int) p->free_scb);
#endif
	/*
	 *  Pause the sequencer so we can play with its registers -
	 *  wait for it to acknowledge the pause.
	 *
	 *  XXX - should the interrupts be left on while doing this?
	 */
	PAUSE_SEQUENCER(p);

	/*
	 *  Save the SCB pointer and put our own pointer in - this
	 *  selects one of the four banks of SCB registers.  Load
	 *  the SCB, then write its pointer into the queue in FIFO
	 *  and restore the saved SCB pointer.
	 */
	old_scbptr = inb(O_SCBPTR(p->base));
	outb(scb->position, O_SCBPTR(p->base));
	
	aic7xxx_putscb(p->base, scb);

	outb(scb->position, O_QINFIFO(p->base));
	outb(old_scbptr, O_SCBPTR(p->base));

	/*
	 *  Make sure the Scsi_Cmnd pointer is saved, the struct it
	 *  points to is set up properly, and the parity error flag
	 *  is reset, then unpause the sequencer and watch the fun
	 *  begin.
	 */
	cmd->scsi_done = fn;
	aic7xxx_parity(cmd) = DID_OK;
	aic7xxx_status(cmd) = 0;

	UNPAUSE_SEQUENCER(p);
	restore_flags(flags);
	return(0);
}

/* return values from aic7xxx_kill */

enum k_state {
	k_ok,				/* scb found and message sent */
	k_busy,				/* message already present */
	k_absent,			/* couldn't locate scb */
	k_disconnect,			/* scb found, but disconnected */
};

/*
 *  This must be called with interrupts disabled - it's going to
 *  be messing around with the host data, and an interrupt being
 *  fielded in the middle could get ugly.
 *
 *  Since so much of the abort and reset code is shared, this
 *  function performs more magic than it really should.  If the
 *  command completes ok, then it will call scsi_done with the
 *  result code passed in.  The unpause parameter controls whether
 *  or not the sequencer gets unpaused - the reset function, for
 *  instance, may want to do something more aggressive.
 *
 *  Note that the command is checked for in our SCB_array first
 *  before the sequencer is paused, so if k_absent is returned,
 *  then the sequencer is NOT paused.
 */

static
enum k_state aic7xxx_kill(Scsi_Cmnd *cmd, unsigned char message,
			  unsigned int result, int unpause)
{
	struct aic7xxx_host *p;
	struct aic7xxx_scb *scb;
	int i, active_scb, found, queued;
	unsigned char scbsave[AIC7XXX_MAXSCB];
	unsigned char flags;
	enum k_state status;

	p = (struct aic7xxx_host *)cmd->host->hostdata;
	scb = &p->scb_array[aic7xxx_position(cmd)];

#if 0
  printk ("aic7xxx_kill: In the kill function...\n");
#endif
	PAUSE_SEQUENCER(p);

	/*
	 *  Case 1: In the QINFIFO
	 *
	 *  This is the best case, really.  Check to see if the
	 *  command is still in the sequencer's input queue.  If
	 *  so, simply remove it.  Reload the queue afterward.
	 */
	queued = inb(O_QINCNT(p->base));
	
	for (i = found = 0; i < queued; i++) {
		scbsave[i] = inb(O_QINFIFO(p->base));

		if (scbsave[i] == scb->position) {
			found = 1;
			i -= 1;
		}
	}

	queued -= found;
	for (i = 0; i < queued; i++)
		outb(scbsave[i], O_QINFIFO(p->base));

	if (found)
	{
	  status = k_ok;
	  goto complete;
	}

	active_scb = inb(O_SCBPTR(p->base));
	/*
	 *  Case 2: Not the active command
	 *
	 *  Check the current SCB bank.  If it's not the one belonging
	 *  to the command we want to kill, select the scb we want to
	 *  abort and turn off the disconnected bit.  The driver will
	 *  then abort the command and notify us of the abort.
	 */
	if (active_scb != scb->position)
	{
	  int scb_control;
	  outb (scb->position, O_SCBPTR(p->base));
	  scb_control = inb (O_SCBARRAY(p->base));
	  scb_control = scb_control & ~SCB_DIS;
	  outb (scb_control, O_SCBARRAY(p->base));
	  outb (active_scb, O_SCBPTR(p->base));
	  status = k_disconnect;
	  goto complete;
	}

	/*
	 *  Presumably at this point our target command is active.  Check
	 *  to see if there's a message already in effect.  If not, place
	 *  our message in and assert ATN so the target goes into MESSAGE
	 *  OUT phase.
	 */
	flags = inb(HA_FLAGS(p->base));
	if (flags & ACTIVE_MSG)
	{
		/*
		 *  If there is a message in progress, reset the bus
		 *  and have all devices renegotiate.
		 */
		if (cmd->lun & 0x8)
		{
		  p->needsdtr = p->needsdtr_copy & 0xff00;
		  p->sdtr_pending = p->sdtr_pending & 0x00ff;
		  outb (0, HA_ACTIVE1(p->base));
		}
		else
		{
		  if (p->bus_type == T_WIDE)
		  {
	   	    p->needsdtr = p->needsdtr_copy;
		    p->needwdtr = p->needwdtr_copy;
		    p->sdtr_pending = 0;
		    p->wdtr_pending = 0;
		    outb(0, HA_ACTIVE0(p->base));
		    outb(0, HA_ACTIVE1(p->base));
		  }
		  else
		  {
		    p->needsdtr = p->needsdtr_copy & 0x00ff;
		    p->sdtr_pending = p->sdtr_pending & 0xff00;
		    outb (0, HA_ACTIVE0(p->base));
		  }
		}
		/* Reset the bus. */
		outb (SCSIRSTO, O_SCSISEQ(p->base));
		udelay(1000);
		outb(0, O_SCSISEQ(p->base));
		aic7xxx_delay (15);

		status = k_busy;
		goto complete;
	}

	outb(flags | ACTIVE_MSG, HA_FLAGS(p->base));  /* active message */
	outb(1, HA_MSG_LEN(p->base));			/* length = 1 */
	outb(message, HA_MSG_START(p->base));		/* message body */

	/*
	 *  Assert ATN.  Use the value of SCSISIGO saved by the
	 *  sequencer code so we don't alter its contents radically
	 *  in the middle of something critical.
	 */
	outb(inb(HA_SIGSTATE(p->base)) | 0x10, O_SCSISIGO(p->base));

	status = k_ok;

	/*
	 *  The command has been killed.  Do the bookkeeping, unpause
	 *  the sequencer, and notify the higher-level SCSI code.
	 */
complete:
	if (unpause)
		UNPAUSE_SEQUENCER(p);

	/*
	 *  Mark the scb as free and clear the scbs command pointer.
	 *  Add the scb to the head of the free list being careful
	 *  to preserve the next pointers.
	 */
	scb->state = SCB_FREE;		/* mark the scb as free */
	scb->cmd = NULL;		/* clear the command pointer */
	scb->next = p->free_scb;	/* preserve next pointer */
	p->free_scb = scb;		/* add at head of free list */
	cmd->result = cmd->result << 16;
	cmd->scsi_done(cmd);
	return(status);
}

int aic7xxx_abort(Scsi_Cmnd *cmd)
{
	int rv;
	long flags;

	save_flags(flags);
	cli();

	switch (aic7xxx_kill(cmd, ABORT, DID_ABORT, !0)) {
	    case k_ok:		rv = SCSI_ABORT_SUCCESS;	break;
	    case k_busy:	rv = SCSI_ABORT_BUSY;		break;
	    case k_absent:	rv = SCSI_ABORT_NOT_RUNNING;	break;
	    case k_disconnect:	rv = SCSI_ABORT_SNOOZE;		break;
	    default:
		panic("aic7xxx_do_abort: internal error\n");
	}

	restore_flags(flags);
	return(rv);
}

/*
 *  Resetting the bus always succeeds - is has to, otherwise the
 *  kernel will panic!  Try a surgical technique - sending a BUS
 *  DEVICE RESET message - on the offending target before pulling
 *  the SCSI bus reset line.
 */

int aic7xxx_reset(Scsi_Cmnd *cmd)
{
	long flags;
	struct aic7xxx_host *p;

	p = (struct aic7xxx_host *)cmd->host->hostdata;
	save_flags(flags);
	cli();

	switch (aic7xxx_kill(cmd, BUS_DEVICE_RESET, DID_RESET, 0)) {

	    case k_ok:
		/*
		 *  The RESET message was sent to the target
		 *  with no problems.  Flag that target as
		 *  needing a SDTR negotiation on the next
		 *  connection and restart the sequencer.
		 */
		p->needsdtr = p->needsdtr & (1 << cmd->target);
		UNPAUSE_SEQUENCER(p);
		break;

	    case k_absent:
		/*
		 *  The sequencer will not be paused if aic7xxx_kill()
		 *  couldn't find the command.
		 */
		PAUSE_SEQUENCER(p);
		/* falls through */

	    case k_busy:
		cmd->result = DID_RESET << 16;	/* return reset code */
		cmd->scsi_done(cmd);
		break;

	    case k_disconnect:
		/*
		 *  Do a hard reset of the SCSI bus.  According to the
		 *  SCSI-2 draft specification, reset has to be asserted
		 *  for at least 25us.  I'm invoking the kernel delay
		 *  function for 30us since I'm not totally trusting of
		 *  the busy loop timing.
		 *
		 *  XXX - I'm not convinced this works.  I tried resetting
		 *	  the bus before, trying to get the devices on the
		 *	  bus to revert to asynchronous transfer, and it
		 *	  never seemed to work.
		 */
		debug("aic7xxx: attempting to reset scsi bus and card\n");

		outb(SCSIRSTO, O_SCSISEQ(p->base));
		udelay(1000);
		outb(0, O_SCSISEQ(p->base));
		aic7xxx_delay (15);

		UNPAUSE_SEQUENCER(p);

		/*
		 *  Locate the command and return a "reset" status
		 *  for it.  This is not completely correct and will
		 *  probably return to haunt me later.
		 */
		cmd->result = DID_RESET << 16;	/* return reset code */
		cmd->scsi_done(cmd);
		break;

	    default:
		panic("aic7xxx_reset: internal error\n");
	}

	restore_flags(flags);
	return(SCSI_RESET_SUCCESS);
}

int aic7xxx_biosparam(Disk *disk, int devno, int geom[])
{
	int heads, sectors, cylinders;
	struct aic7xxx_host *p;
	
	p = (struct aic7xxx_host *)disk->device->host->hostdata;

	/*
	 *  XXX - if I could portably find the card's configuration
	 *	  information, then this could be autodetected instead
	 *	  of left to a boot-time switch.
	 */
	heads = 64;
	sectors = 32;
	cylinders = disk->capacity / (heads * sectors);

	if (p->extended && cylinders > 1024) {
		heads = 255;
		sectors = 63;
		cylinders = disk->capacity / (255 * 63);
	}

	geom[0] = heads;
	geom[1] = sectors;
	geom[2] = cylinders;

	return(0);
}


